<?php
namespace Teknisa\SUP\Modules\Cadastros\Services;

use PHPExcel_IOFactory;
use Zeedhi\Framework\DTO;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Framework\DB\StoredProcedure\Param;
use Zeedhi\Framework\DB\StoredProcedure\StoredProcedure;

class ConfigAlcadasSheetImporter 
{
    const SHEETS_FOLDER = __DIR__ . "/../Temp/";
    const SHEET_TEMPLATE = __DIR__ . "/../Temp/Modelo.xlsx";
    
    const COLUMNS_MAP = array(
        "NMFILIAL" => 
            array(
                "index"     => 0,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "ns1:Instalacion"
            ),
        "NSTITULAR" => 
            array(
                "index"     => 1,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Titular"
            ),
        "NSDIRECCION" => 
            array(
                "index"     => 2,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Direccion"
            ),
        "NSPOBLACION" => 
            array(
                "index"     => 3,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Poblacion"
            ),
        "NSCODIGOPOSTAL" => 
            array(
                "index"     => 4,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:CodigoPostal"
            ),
        "NSTELEFONO" => 
            array(
                "index"     => 5,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Telefono"
            ),
        "NSEMAIL" => 
            array(
                "index"     => 6,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Email"
            ),
        "CDCENTROCUS" => 
            array(
                "index"     => 7,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "ns1:CodigoCentroCoste"
            ),
        "NSBARRIO" => 
            array(
                "index"     => 8,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Barrio"
            ),
        "NSCERRADO" => 
            array(
                "index"     => 9,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Cerrado"
            ),
        "NSREGION" => 
            array(
                "index"     => 10,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Region"
            ),
        "NSFECHAAPERTURA" => 
            array(
                "index"     => 11,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:FechaApertura"
            ),
        "NMAPROVN1" => 
            array(
                "index"     => 12,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Autorizador"
            ),
        "DSEMAILAPROVN1" => 
            array(
                "index"     => 13,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:EmailAutorizador"
            ),
        "NMAPROVN2" => 
            array(
                "index"     => 14,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Autorizador2"
            ),
        "DSEMAILAPROVN2" => 
            array(
                "index"     => 15,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:EmailAutorizador2"
            ),
        "NSNIVEL1" => 
            array(
                "index"     => 16,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "ns1:Nivel1"
            ),
        "DSSEGMENTO" => 
            array(
                "index"     => 17,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "ns1:Nivel2"
            ),
        "DSSUPERINTEND" => 
            array(
                "index"     => 18,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "ns1:Nivel3"
            ),
        "DSDIRETORIA" => 
            array(
                "index"     => 19,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "ns1:Nivel4"
            ),
        "DSREGIONAL" => 
            array(
                "index"     => 20,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "ns1:Nivel5"
            ),
        "DSGOP" => 
            array(
                "index"     => 21,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "ns1:Nivel6"
            )
    );
    
    protected $errors = array();
    protected $errorMessages = array();
    protected $newCodeSC;

    public function __construct($entityManager, $environment, $newCodeSC) 
    {
        $this->entityManager = $entityManager;
        $this->connection = $this->entityManager->getConnection();
        $this->environment = $environment;
        $this->newCodeSC = $newCodeSC;
    }
    
    /**
     * Start a transaction on database
     * 
     * @return void
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function beginTransaction()
    {
        $this->entityManager->beginTransaction();
    }
    
    /**
     * Commit a transaction on database
     * 
     * @return void
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function commitTransaction()
    {
        $this->entityManager->commit();
    }
    
    /**
     * Rollback a transaction on database
     * 
     * @return void
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function rollbackTransaction()
    {
        $this->entityManager->rollBack();
    }
    
    private function generateNewCode($tableName) {
        return $this->newCodeSC->getCodeByTableName($tableName, $this->environment->getNrOrg());
    }
    
    public function downloadTemplate(DTO\Request $request, DTO\Response $response) 
    {
        $name = "Modelo.xls";
        $template = self::SHEET_TEMPLATE;
        
        header("Accept-Ranges: none");
        header("Content-Type: " . "xlsx");
        header("Content-Disposition: attachment; filename=" . $name);
        
        readfile($template);
        
        exit;
    }
    
    public function createTemporaryFile($sheetImport) 
    {
        if ( isset($sheetImport) &&  is_array($sheetImport) &&  !empty($sheetImport) 
             && isset($sheetImport[0]["b64File"]) && isset($sheetImport[0]["name"]) ) {
            $sheetB64 = $sheetImport[0]["b64File"];
            $sheetName = $sheetImport[0]["name"];
            $sheet = $this->decodeFromBase64($sheetB64);
            $filePath = self::SHEETS_FOLDER . "tmp-" . $sheetName;
            file_put_contents($filePath, $sheet);
            chmod($filePath, 0644);
            return $filePath;
        } else {
            throw new \Exception ("sheetUploadFailure");
        }
    }
    
    private function decodeFromBase64($sheet) 
    {
        $b64File = explode(",", $sheet);
        $base64_decode = base64_decode($b64File[1], true);

        return $base64_decode;
    }
    
    public function loadSheet($sheetPath) 
    {
        try {
            $objPHPExcel = PHPExcel_IOFactory::load($sheetPath);
            return $objPHPExcel->getActiveSheet();
        } catch( \Excention $e) {
            throw new \Exception("invalidSheet");
            
        }
    }
    
    public function parseSheet($worksheet) 
    {
        try {
            $rows = array();
            
            foreach ($worksheet->getRowIterator() AS $row) {
                $cellIterator = $row->getCellIterator();
                $cellIterator->resetEnd("V");
                $cellIterator->setIterateOnlyExistingCells(FALSE); // This loops through all cells,
                $cells = [];
                $hasNotNullCell = false;
                
                foreach ($cellIterator as $cell) {
                    $value = $cell->getValue();
                    $hasNotNullCell = $hasNotNullCell || $value;
                    $cells[] = $value;
                }
                
                if ($hasNotNullCell) { // At least one cell with value
                    $rows[] = $cells;
                }
            }
            
            if (((is_array($rows) && !is_null($rows)) ? count($rows) : 0) < 2) {
                throw new \Exception("emptySheet");
            }
        
            return $rows;
        } catch( \Excention $e) {
            throw $e;
        }
    }
    
    public function validateSheetColumns($sheet) 
    {
        if (isset($sheet[0])) {
            $header = $sheet[0];
            
            $count = (is_array($header) && !is_null($header)) ? count($header) : 0;

            if ($count == 22) {
                foreach (self::COLUMNS_MAP as $column) {
                    if ($header[$column["index"]] != $column["label"]) {
                        throw new \Exception("invalidSheetColumnsLabels");
                    }
                }
            }
        } else {
            throw new \Exception("invalidSheet");
        }
    }
    
    public function validateSheetRows(&$sheet, $response) 
    {
        foreach ($sheet as $key => &$row) {
            if ($key != 0) { // Skip Header Row
                $this->validateEmptyValues($row, $key);
                $this->validateInvalidValues($row, $key);
                
            }
        }
        
        if (!empty($this->errors)) {
            $this->decodeErrorMessages();
            $response->addDataSet(new DataSet("errors", $this->errorMessages));
            throw new \Exception("sheetErrors", 1);
        }
    }
    
    private function validateEmptyValues(&$row, $rowNum) 
    {
        foreach (self::COLUMNS_MAP as $columnName => $column) {
            if (is_null($row[$column["index"]]) && !$column["nullable"]) {
                $this->addError($rowNum, "emptyCell", $columnName);
            }
        }
    }
    
    private function validateInvalidValues(&$row, $rowNum) 
    {
        foreach (self::COLUMNS_MAP as $columnName => $column) {
            if (strlen($row[$column["index"]]) != 0) {
                if (isset($column["length"]) && strlen($row[$column["index"]]) != $column["length"]) {
                    if (strlen($row[$column["index"]]) < $column["length"] && isset($column["zerofill"]) && $column["zerofill"]) {
                        $row[$column["index"]] = str_pad($row[$column["index"]], $column["length"], "0", STR_PAD_LEFT);
                    } else {
                        $this->addError($rowNum, "invalidLength", $columnName); 
                    }
                } else if (isset($column["maxLength"]) && strlen($row[$column["index"]]) > $column["maxLength"]) {
                    $this->addError($rowNum, "invalidMaxLength", $columnName); 
                }
            }
        }
    }
    
    private function decodeErrorMessages() 
    {
        $tab = "          ";
        
        foreach ($this->errors as $rowNum => $error) {
            $message = "Linha " . ($rowNum + 1) . ":";
            array_push($this->errorMessages, $message);
            
            if (isset($error["emptyCell"])) {
                foreach (($error["emptyCell"]) as $column) {
                    $message = $tab . "O \"" . self::COLUMNS_MAP[$column]["label"] . "\" deve ser informado.";
                    array_push($this->errorMessages, $message);
                }
            }

            if (isset($error["invalidLength"])) {
                foreach (($error["invalidLength"]) as $column) {
                    $message = $tab . "O \"" . self::COLUMNS_MAP[$column]["label"] . "\" deve possuir exatamente ". self::COLUMNS_MAP[$column]["length"] . " caracteres.";
                    array_push($this->errorMessages, $message);
                }
            }

            if (isset($error["invalidMaxLength"])) {
                foreach (($error["invalidMaxLength"]) as $column) {
                    $message = $tab . "O \"" . self::COLUMNS_MAP[$column]["label"] . "\" deve possuir no máximo ". self::COLUMNS_MAP[$column]["maxLength"] . " caracteres.";
                    array_push($this->errorMessages, $message);
                }
            }
        }
    }
    
    private function addError($rowNum, $errorType, $error) 
    {
        if (!isset($this->errors[$rowNum])) {
            $this->errors[$rowNum] = array();
        }
        
        if (!isset($this->errors[$rowNum][$errorType])) {
            $this->errors[$rowNum][$errorType] = array();
        }
        
        array_push($this->errors[$rowNum][$errorType], $error);
    }
    
    public function validateAndInsertData($sheet, $response)
    {
        $dataArrays = self::getDataArray($sheet);
        $sequential = self::insertImportedData($dataArrays);
        return $sequential;
    }
    
    private function getDataArray($sheet)
    {
        $dataArrays = [];
        
        foreach ($sheet as $ind => $data) {
            if($ind != 0){
                $dataArrays[$ind] = $data;
            }
        }
        
        return $dataArrays;
    }
    
    private function getInsertQuery()
    {
        $sql = "INSERT INTO INTG_HIERARQ(
                   NRSEQIMPORT,
                   NMFILIAL,
                   CDCENTROCUS,
                   NMAPROVN1,
                   DSEMAILAPROVN1,
                   NMAPROVN2,
                   DSEMAILAPROVN2,
                   DSSEGMENTO,
                   DSSUPERINTEND,
                   DSDIRETORIA,
                   DSREGIONAL,
                   DSGOP,
                   IDIMPORTADO,
                   NRLINHAARQ
                )
                VALUES
                (
                    :P_NRSEQIMPORT,
                    :P_NMFILIAL,
                    :P_CDCENTROCUS,
                    :P_NMAPROVN1,
                    :P_DSEMAILAPROVN1,
                    :P_NMAPROVN2,
                    :P_DSEMAILAPROVN2,
                    :P_DSSEGMENTO,
                    :P_DSSUPERINTEND,
                    :P_DSDIRETORIA,
                    :P_DSREGIONAL,
                    :P_DSGOP,
                    :P_IDIMPORTADO,
                    :P_NRLINHAARQ
                )";
                
        return $sql;
    }
    
    private function insertImportedData($dataArrays)
    {
        $sql = $this->getInsertQuery();
        $newCode = $this->generateNewCode("INTG_HIERARQ");

        foreach ($dataArrays as $index => $dataArray) {
            $data = $this->removeSpecialCharacter($dataArray);
            
            $param = array();
            $param["P_NRSEQIMPORT"] = $newCode;
            $param["P_NMFILIAL"] = (string)$data[0];
            $param["P_CDCENTROCUS"] = (string)$data[7];
            $param["P_NMAPROVN1"] = (string)$data[12];
            $param["P_DSEMAILAPROVN1"] = (string)$data[13];
            $param["P_NMAPROVN2"] = (string)$data[14];
            $param["P_DSEMAILAPROVN2"] = (string)$data[15];
            $param["P_DSSEGMENTO"] = (string)$data[17];
            $param["P_DSSUPERINTEND"] = (string)$data[18];
            $param["P_DSDIRETORIA"] = (string)$data[19];
            $param["P_DSREGIONAL"] = (string)$data[20];
            $param["P_DSGOP"] = (string)$data[21];
            $param["P_IDIMPORTADO"] = 'N';
            $param["P_NRLINHAARQ"] = $index;
            
            try {
                $std = $this->connection->prepare($sql);
                $std->execute($param);
            } catch (Exception $e){
                throw new \Exception("invalidSheet");
            }
        }
        
        return $newCode;
    }
    
    private function removeSpecialCharacter($array)
    {
        foreach($array as $index => $value) {
            $array[$index] = str_replace("&", "", $value);
        }
        
        return $array;
    }
    
    public function removeTemporaryFile($filePath) 
    {
        unlink($filePath);
    }
    
    public function startImportProcedure($sequential)
    {
        try{
            $procedure = new StoredProcedure($this->connection, 'USR_IMPORT_HIERARQ');
            $procedure->addParam(new Param('P_NRSEQIMPORT', Param::PARAM_INPUT, $sequential, Param::PARAM_TYPE_STR));
            $procedure->addParam(new Param('P_CDOPERADOR', Param::PARAM_INPUT, $this->environment->getCdOperador(), Param::PARAM_TYPE_STR));
            
            $procedure->execute();
        }catch(\Exception $e){
            return $e;
        }
    }
    
    public function checkForErrors($sequential)
    {
        $sql = "SELECT
                    *
                FROM
                    LOG_INTG_HIERARQ
                WHERE
                    NRSEQIMPORT = :P_NRSEQIMPORT";
        
        $param = array(
            "P_NRSEQIMPORT" => $sequential
        );
        
        $std = $this->connection->prepare($sql);
        $std->execute($param);
        return $std->fetchAll();
    }
}